unit uPlugin;
{$INCLUDE 'VirusTotal.inc'}

interface

implementation

uses
  Windows, SysUtils, Classes,
  {$INCLUDE 'DelphiFarUnits.inc'} ,
  DCPcrypt2, DCPsha256,
  {$IFDEF DEBUG_LOGS} uLog, {$ENDIF}
  uOptions, uOptionsUI,
  uReportUI,
  uVirusTotal,
  uSystem, uConsts, uMessages;

type
  TPlugin = class(TFarPlugin)
  private
    fPluginInfoAvailable: boolean;
    fPluginInfo: TPluginInfo;
    fOptions: TOptions;
    fProgress: TFarProgress;
    function GetOptions: TOptions;
  protected
    function PluginID: TGUID; override;
    function GetPluginTitle(out Description, Author: string): string; override;
    function GetPluginInfo(out Info: TPluginInfo): boolean; override;
    function Execute(OpenFrom: DWORD; GUID: PGUID; Data: TIntPtr): THandle; override;
    function GetCommonString(ID: TFarPlugin.TCommonString): string; override;
    procedure DoLog(const Msg: string); override;
    function PluginConfig(GUID: PGUID; MenuItem: TIntPtr): boolean; override;
    function ProcessPanel(const Active: boolean; const FileName: string = ''): boolean;
    function ProcessFileName(const FileName: string; ForceRefresh: boolean): boolean;
    function ProcessStream(const FileName: string; Data: TStream; ForceRefresh: boolean): boolean;
    function ProcessStreamWithoutDisplay(Data: TStream; ForceRefresh: boolean): IVirusTotalReport;
    function ProcessMacroCall(const Params: array of Variant): boolean;
    procedure ShowReport(const FileName: string; const Report: IVirusTotalReport; Data: TStream);
    procedure ReloadReport(Sender: TReportDialog; var NewReport: IVirusTotalReport);
    procedure RefreshOptions;
    property Options: TOptions read GetOptions;
    procedure ShowProgress(const Description, Progress: string; FullRefresh: boolean = False); overload;
    procedure ShowProgress(const Description: string; const Position, Max: int64; FullRefresh: boolean = False); overload;
    procedure ShowProgress(const Description: string; FullRefresh: boolean = False); overload;
    procedure HideProgress;
    procedure EscapePressedDuringProgress(Sender: TFarProgress; var DoAbort: boolean);
    procedure WaitWithProgress(const Seconds: integer; const ProgressMessage: string);
    function IsProgress: boolean;
    function GetVirusTotalReport(const ScanID: string): IVirusTotalReport;
    function GetVirusTotalScan(Data: TStream): IVirusTotalScan;
    function GetVirusTotalRescan(const ScanID: string): IVirusTotalScan;
    function HandleVirusTotalError(const E: Exception): boolean;
  public
    constructor Create;
    destructor Destroy; override;
  end;

const
  PLUGIN_TITLE = 'VirusTotal';
  PLUGIN_DESCRIPTION = 'VirusTotal';
  PLUGIN_AUTHOR = 'Pepak, https://www.pepak.net';

const
  BLOCK_SIZE = 4*1024*1024;

{ TPlugin }

constructor TPlugin.Create;
begin
  inherited;
  fPluginInfoAvailable := False;
  fOptions := nil;
  fProgress := nil;
end;

destructor TPlugin.Destroy;
begin
  FreeAndNil(fOptions);
  FreeAndNil(fProgress);
  if fPluginInfoAvailable then begin
    DeletePluginInfo(fPluginInfo);
  end;
  inherited;
end;

procedure TPlugin.DoLog(const Msg: string);
begin
  {$IFDEF DEBUG_LOGS}
  uLog.Log(Msg);
  {$ENDIF}
end;

procedure TPlugin.EscapePressedDuringProgress(Sender: TFarProgress; var DoAbort: boolean);
begin
  DoAbort := ShowMessage([GetMsg(MTitle), GetMsg(MDoYouWantToCancel), GetMsg(MYes), GetMsg(MNo)], FMSG_WARNING, 2) = 0;
end;

function TPlugin.Execute(OpenFrom: DWORD; GUID: PGUID; Data: TIntPtr): THandle;
var
  CommandLine: string;
begin
  Result := {$IFDEF FAR3_UP} 0 {$ELSE} INVALID_HANDLE_VALUE {$ENDIF} ;
  RefreshOptions;
  {$IFDEF FAR3_UP}
  if OpenFrom = OPEN_FROMMACRO_ then
  begin
    ProcessMacroCall(TFarMacro.MacroParamsToDelphi(Data));
    Exit;
  end;
  {$ENDIF}
  if OpenFrom = OPEN_COMMANDLINE then
  begin
    CommandLine := CopyString( {$IFDEF FAR3_UP} POpenCommandLineInfo(Data)^.CommandLine {$ELSE} PFarChar(Data) {$ENDIF} );
    CommandLine := AnsiDequotedStr(Trim(CommandLine), '"');
    if CommandLine <> '' then
    begin
      if IsRelativePath(CommandLine) then
        ProcessPanel(True, CommandLine)
      else
        ProcessFileName(CommandLine, False);
    end;
    Exit;
  end;
  if (OpenFrom = OPEN_FILEPANEL) or (OpenFrom = OPEN_PLUGINSMENU) then
  begin
    ProcessPanel(True);
  end;
end;

function TPlugin.GetCommonString(ID: TFarPlugin.TCommonString): string;
begin
  case ID of
    TFarPlugin.TCommonString.MTitle:
      Result := GetMsg(MTitle);
    TFarPlugin.TCommonString.MExceptionOccurredIn:
      Result := GetMsg(MExceptionOccurredIn);
    TFarPlugin.TCommonString.MOk:
      Result := GetMsg(MOk);
    TFarPlugin.TCommonString.MCancel:
      Result := GetMsg(MCancel);
    TFarPlugin.TCommonString.MYes:
      Result := GetMsg(MYes);
    TFarPlugin.TCommonString.MNo:
      Result := GetMsg(MNo);
    else
      Result := inherited GetCommonString(ID);
  end;
end;

function TPlugin.GetOptions: TOptions;
begin
  if fOptions = nil then
  begin
    fOptions := TOptions.Create;
    fOptions.Load;
  end;
  Result := fOptions;
end;

function TPlugin.GetPluginInfo(out Info: TPluginInfo): boolean;
var
  fMainMenu: TFarPlugin.TFarPluginMenu;
  fConfigMenu: TFarPlugin.TFarPluginMenu;
  fDiskMenu: TFarPlugin.TFarPluginMenu;
begin
  if not fPluginInfoAvailable then begin
    SetLength(fMainMenu, 1);
    fMainMenu[0].Title := GetMsg(MTitle);
    fMainMenu[0].ID := MENU_PANEL_GUID;
    SetLength(fConfigMenu, 1);
    fConfigMenu[0].Title := GetMsg(MTitle);
    fConfigMenu[0].ID := MENU_CONFIG_GUID;
    SetLength(fDiskMenu, 0);
    Self.fPluginInfo := CreatePluginInfo(
      {$IFDEF FAR3_UP} PF_NONE {$ELSE} 0 {$ENDIF} ,
      fMainMenu, fConfigMenu, fDiskMenu,
      Options.CommandLinePrefix
    );
    fPluginInfoAvailable := True;
  end;
  Info := Self.fPluginInfo;
  Result := True;
end;

function TPlugin.GetPluginTitle(out Description, Author: string): string;
begin
  Result := PLUGIN_TITLE;
  Description := PLUGIN_DESCRIPTION;
  Author := PLUGIN_AUTHOR;
end;

function TPlugin.GetVirusTotalReport(const ScanID: string): IVirusTotalReport;
var
  VirusTotal: IVirusTotal;
  NeedHideProgress: boolean;
  Report: IVirusTotalReport;
begin
  NeedHideProgress := not IsProgress;
  try
    ShowProgress(GetMsg(MReadingReport), True);
    VirusTotal := GetVirusTotalInstance(Options.ApiKey);
    repeat
      try
        Report := VirusTotal.GetReport(ScanID);
      except
        on E: Exception do
        begin
          if HandleVirusTotalError(E) then
            Report := nil
          else
            Abort;
        end;
      end;
      if (Report = nil) or (Report.ResponseResult = vtPending) then
      begin
        WaitWithProgress(Options.DelayBetweenReportChecks, GetMsg(MReadingReport));
      end
      else
      begin
        case Report.ResponseResult of
          vtOK:
            begin
              Result := Report;
              Break;
            end;
          vtNotFound:
            begin
              Result := nil;
              Break;
            end;
          else
            begin
              ShowWarning(GetMsg(MUnknownServerResponse));
              Abort;
            end;
        end;
      end;
    until False;
  finally
    if NeedHideProgress then
      HideProgress;
  end;
end;

function TPlugin.GetVirusTotalRescan(const ScanID: string): IVirusTotalScan;
var
  NeedHideProgress: boolean;
  VirusTotal: IVirusTotal;
  Scan: IVirusTotalScan;
begin
  NeedHideProgress := not IsProgress;
  try
    ShowProgress(GetMsg(MRescanningFile), True);
    VirusTotal := GetVirusTotalInstance(Options.ApiKey);
    repeat
      try
        Scan := VirusTotal.Rescan(ScanID);
      except
        on E: Exception do
        begin
          if HandleVirusTotalError(E) then
            Scan := nil
          else
            Abort;
        end;
      end;
      if (Scan = nil) or (Scan.ResponseResult = vtPending) then
      begin
        WaitWithProgress(Options.DelayBetweenReportChecks, GetMsg(MRescanningFile));
      end
      else
      begin
        case Scan.ResponseResult of
          vtOK:
            begin
              Result := Scan;
              Break;
            end;
          vtNotFound:
            begin
              Result := nil;
              Break;
            end;
          else
            begin
              ShowWarning(GetMsg(MUnknownServerResponse));
              Abort;
            end;
        end;
      end;
    until False;
  finally
    if NeedHideProgress then
      HideProgress;
  end;
end;

function TPlugin.GetVirusTotalScan(Data: TStream): IVirusTotalScan;
var
  NeedHideProgress: boolean;
  VirusTotal: IVirusTotal;
  Scan: IVirusTotalScan;
begin
  NeedHideProgress := not IsProgress;
  try
    ShowProgress(GetMsg(MUploadingFile), True);
    VirusTotal := GetVirusTotalInstance(Options.ApiKey);
    VirusTotal.OnUploadProgress :=
      procedure(const Sender: IVirusTotal; const BytesTransferred, TotalBytes: int64)
      begin
        ShowProgress(GetMsg(MUploadingFile), BytesTransferred, TotalBytes);
      end;
    repeat
      try
        Data.Position := 0;
        Scan := VirusTotal.ScanFile(Data);
      except
        on E: Exception do
        begin
          if HandleVirusTotalError(E) then
            Scan := nil
          else
            Abort;
        end;
      end;
      if (Scan = nil) or (Scan.ResponseResult = vtPending) then
      begin
        WaitWithProgress(Options.DelayBetweenReportChecks, GetMsg(MUploadingFile));
      end
      else
      begin
        case Scan.ResponseResult of
          vtOK:
            begin
              Result := Scan;
              Break;
            end;
          vtNotFound:
            begin
              Result := nil;
              Break;
            end;
          else
            begin
              ShowWarning(GetMsg(MUnknownServerResponse));
              Abort;
            end;
        end;
      end;
    until False;
  finally
    if NeedHideProgress then
      HideProgress;
  end;
end;

function TPlugin.HandleVirusTotalError(const E: Exception): boolean;
begin
  if E = nil then
    Result := True
  else if E is EVirusTotalThrottlingError then
  begin
    Result := ShowMessage([GetMsg(MTitle), GetMsg(MRateExceeded), GetMsg(MRetry), GetMsg(MAbort)], FMSG_WARNING, 2) = 0;
  end
  else if E is EVirusTotalPrivilegesError then
  begin
    ShowWarning(GetMsg(MForbidden));
    Result := False;
  end
  else
  begin
    ShowWarning(E.Message);
    Result := False;
  end;
end;

procedure TPlugin.HideProgress;
begin
  FreeAndNil(fProgress);
end;

function TPlugin.IsProgress: boolean;
begin
  Result := (fProgress <> nil);
end;

function TPlugin.PluginConfig(GUID: PGUID; MenuItem: TIntPtr): boolean;
var
  Dialog: TOptionsDialog;
begin
  Result := False;
  RefreshOptions;
  Dialog := TOptionsDialog.Create;
  try
    Dialog.Options := Options;
    if Dialog.Execute then
    begin
      Options.Save;
      Result := True;
    end;
  finally
    FreeAndNil(Dialog);
  end;
end;

function TPlugin.PluginID: TGUID;
begin
  Result := PLUGIN_GUID;
end;

function TPlugin.ProcessFileName(const FileName: string; ForceRefresh: boolean): boolean;
var
  NeedHideProgress: boolean;
  SR: TSearchRec;
  Stream: TMemoryStream;
  SourceStream: TFileStream;
  n: int64;
begin
  Result := False;
  if FileName = '' then
    Exit;
  NeedHideProgress := not IsProgress;
  try
    ShowProgress(GetMsg(MSettingUp), True);
    // Verify the validity of input data
    if Options.ApiKey = '' then
    begin
      ShowWarning(GetMsg(MNoApiKey));
      Exit;
    end;
    if FindFirst(FileName, 0, SR) <> 0 then
    begin
      ShowMessage([GetMsg(MTitle), GetMsg(MFileNotFound), FileName, GetMsg(MOk)], FMSG_WARNING, 1);
      Exit;
    end;
    try
      if (SR.Attr and (faDirectory)) <> 0 then
        Exit;
      if SR.Size <= 0 then
        Exit;
      if SR.Size > Options.MaxFileSize then
      begin
        ShowWarning(Format(GetMsg(MFileTooLarge), [Options.MaxFileSize]));
        Exit;
      end;
      // Load the file into memory
      ShowProgress(GetMsg(MReadingFile), 0, 1, True);
      Stream := TMemoryStream.Create;
      try
        SourceStream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
        try
          while SourceStream.Position < SourceStream.Size do
          begin
            ShowProgress(GetMsg(MReadingFile), SourceStream.Position, SourceStream.Size);
            n := SourceStream.Size - SourceStream.Position;
            if n > BLOCK_SIZE then
              n := BLOCK_SIZE;
            Stream.CopyFrom(SourceStream, n);
          end;
        finally
          FreeAndNil(SourceStream);
        end;
        // Process the file
        Result := ProcessStream(FileName, Stream, ForceRefresh);
      finally
        FreeAndNil(Stream);
      end;
    finally
      FindClose(SR);
    end;
  finally
    if NeedHideProgress then
      HideProgress;
  end;
end;

function TPlugin.ProcessMacroCall(const Params: array of Variant): boolean;
var
  FileName: string;
begin
  Result := False;
  if Length(Params) > 0 then
  begin
    FileName := Params[0];
    if (FileName = '') or IsRelativePath(FileName) then
      Result := ProcessPanel(True, FileName)
    else
      Result := ProcessFileName(FileName, False);
  end;
end;

function TPlugin.ProcessPanel(const Active: boolean; const FileName: string): boolean;
var
  Panel: TFarPanel;
  PanelItemIndex: integer;
  FullFileName: string;
  i: integer;
begin
  Result := False;
  Panel := TFarPanel.Create(Active, True);
  try
    FullFileName := '';
    PanelItemIndex := -1;
    if FileName <> '' then
    begin
      if PanelItemIndex < 0 then
      begin
        for i := 0 to Pred(Panel.ItemCount) do
        begin
          if Panel.Items[i].FileName = FileName then
          begin
            PanelItemIndex := i;
            Break;
          end;
        end;
      end;
      if PanelItemIndex < 0 then
      begin
        for i := 0 to Pred(Panel.ItemCount) do
        begin
          if AnsiCompareText(Panel.Items[i].FileName, FileName) = 0 then
          begin
            PanelItemIndex := i;
            Break;
          end;
        end;
      end;
    end
    else
    begin
      PanelItemIndex := Panel.CurrentItemIndex;
    end;
    if (PanelItemIndex >= 0) and (PanelItemIndex < Panel.ItemCount) then
    begin
      {$IFDEF FAR3_UP}
      if Longbool(Panel.Flags and PFLAGS_PLUGIN) then
      begin
        if Longbool(Panel.Flags and PFLAGS_REALNAMES) then
          FullFileName := Panel.Items[PanelItemIndex].FileName;
      end
      else
      {$ENDIF}
      begin
        if Panel.Directory <> '' then
          FullFileName := IncludeTrailingPathDelimiter(Panel.Directory) + Panel.Items[PanelItemIndex].FileName
        else
          FullFileName := Panel.Items[PanelItemIndex].FileName;
      end;
    end;
    if FullFileName <> '' then
      Result := ProcessFileName(FullFileName, False)
    else
      ShowMessage([GetMsg(MTitle), GetMsg(MFileNotFound), FileName, GetMsg(MOk)], FMSG_WARNING, 1);
  finally
    FreeAndNil(Panel);
  end;
end;

function TPlugin.ProcessStream(const FileName: string; Data: TStream; ForceRefresh: boolean): boolean;
var
  Report: IVirusTotalReport;
begin
  Report := ProcessStreamWithoutDisplay(Data, ForceRefresh);
  if Report = nil then
    Abort;
  ShowReport(FileName, Report, Data);
  Result := True;
end;

function TPlugin.ProcessStreamWithoutDisplay(Data: TStream; ForceRefresh: boolean): IVirusTotalReport;
var
  NeedHideProgress: boolean;
  Hash: TDCP_hash;
  HashBytes: array of Byte;
  ScanID: string;
  Scan: IVirusTotalScan;
  i: integer;
  n: int64;
begin
  Result := nil;
  NeedHideProgress := not IsProgress;
  try
    // Calculate the file's hash
    ShowProgress(GetMsg(MHashingFile), 0, 1, True);
    Hash := TDCP_sha256.Create(nil);
    try
      Hash.Init;
      Data.Position := 0;
      while Data.Position < Data.Size do
      begin
        ShowProgress(GetMsg(MHashingFile), Data.Position, Data.Size);
        n := Data.Size - Data.Position;
        if n > BLOCK_SIZE then
          n := BLOCK_SIZE;
        Hash.UpdateStream(Data, n);
      end;
      SetLength(HashBytes, Hash.GetHashSize div 8);
      Hash.Final(HashBytes[0]);
      ScanID := '';
      for i := 0 to Pred(Length(HashBytes)) do
        ScanID := ScanID + IntToHex(HashBytes[i], 2);
    finally
      FreeAndNil(Hash);
    end;
    ScanID := LowerCase(ScanID);
    // Read a previous report, if possible
    if ForceRefresh then
      Result := nil
    else
      Result := GetVirusTotalReport(ScanID);
    // Upload the file
    if Result = nil then
    begin
      Scan := GetVirusTotalScan(Data);
      if Scan = nil then
        Abort;
      if Scan.ResponseResult <> vtOK then
        Abort;
      if Scan.ScanID = '' then
      begin
        ShowWarning(GetMsg(MServerResponseMissingScanID));
        Abort;
      end;
      ScanID := Scan.ScanID;
      // Wait for the initial period
      WaitWithProgress(Options.InitialScanTime, GetMsg(MWaitingForScanResults));
      // Read the report
      Result := GetVirusTotalReport(ScanID);
    end;
    // Interpret the report
  finally
    if NeedHideProgress then
      HideProgress;
  end;
end;

procedure TPlugin.RefreshOptions;
begin
  FreeAndNil(fOptions);
end;

procedure TPlugin.ReloadReport(Sender: TReportDialog; var NewReport: IVirusTotalReport);
var
  NeedHideProgress: boolean;
  Report: IVirusTotalReport;
  ScanID: string;
  Scan: IVirusTotalScan;
begin
  NeedHideProgress := not IsProgress;
  try
    // Rescan API would be better, but I wasn't able to get it to work
    if Options.UseRescanAPI or (Sender.Data = nil) then
    begin
      Scan := GetVirusTotalRescan(Sender.Report.SHA256);
      if Scan = nil then
        Abort;
      if Scan.ResponseResult <> vtOK then
        Abort;
      if Scan.ScanID = '' then
      begin
        ShowWarning(GetMsg(MServerResponseMissingScanID));
        Abort;
      end;
      ScanID := Scan.ScanID;
      // Wait for the initial period
      WaitWithProgress(Options.InitialScanTime, GetMsg(MWaitingForScanResults));
      // Read the report
      Report := GetVirusTotalReport(ScanID);
    end
    else
    begin
      Report := ProcessStreamWithoutDisplay(Sender.Data, True);
    end;
    // Interpret the report
    if Report = nil then
      Abort;
    if NeedHideProgress then
      HideProgress;
    NewReport := Report;
  finally
    if NeedHideProgress then
      HideProgress;
  end;
end;

procedure TPlugin.ShowReport(const FileName: string; const Report: IVirusTotalReport; Data: TStream);
var
  Dialog: TReportDialog;
begin
  if Report = nil then
  begin
    ShowMessage(GetMsg(MReportIsEmpty));
    Exit;
  end;
  Dialog := TReportDialog.Create;
  try
    Dialog.FileName := FileName;
    Dialog.Report := Report;
    Dialog.Data := Data;
    if Options.UseRescanAPI or Assigned(Data) then
      Dialog.OnReloadReport := ReloadReport;
    Dialog.Execute;
  finally
    FreeAndNil(Dialog);
  end;
end;

procedure TPlugin.WaitWithProgress(const Seconds: integer; const ProgressMessage: string);
var
  NeedHideProgress: boolean;
  StartTime, EndTime, CurrentTime: TDateTime;
begin
  NeedHideProgress := not IsProgress;
  try
    ShowProgress(ProgressMessage, 0, 1, True);
    StartTime := Now;
    EndTime := StartTime + Seconds/(24*60*60);
    repeat
      CurrentTime := Now;
      if CurrentTime > EndTime then
        Break;
      Sleep(200);
      ShowProgress(ProgressMessage, Trunc((CurrentTime - StartTime) * (24*60*60) * 10), 10*Seconds);
    until False;
  finally
    if NeedHideProgress then
      HideProgress;
  end;
end;

procedure TPlugin.ShowProgress(const Description, Progress: string; FullRefresh: boolean);
begin
  if fProgress = nil then
  begin
    fProgress := TFarProgress.Create;
    fProgress.CheckForEscape := True;
    fProgress.OnEscapePressed := EscapePressedDuringProgress;
  end;
  fProgress.Show([GetMsg(MTitle), Description, Progress], FullRefresh);
end;

procedure TPlugin.ShowProgress(const Description: string; FullRefresh: boolean);
begin
  ShowProgress(Description, '', FullRefresh);
end;

procedure TPlugin.ShowProgress(const Description: string; const Position, Max: int64; FullRefresh: boolean);
begin
  ShowProgress(Description, fProgress.BuildProgressBar(Position, Max), FullRefresh);
end;

initialization
  TFarPlugin.InitInstance(TPlugin.Create);

end.
